//////////////////////////////////////////////////////////////////////
//       SpeedStep Centrino © Gergely Rózsahegyi (RoGer) 2005       //
//                           roger@beos.hu                          //
//                                                                  //
//                        Driver Source File                        //
//                                                                  //
//                       Created:  2005.05.23                       //
//                                                                  //
//////////////////////////////////////////////////////////////////////

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include "SpeedStep.h"

#define MSR_IA32_MISC_ENABLE 0x1a0
#define MSR_IA32_PERF_STATUS 0x198
#define MSR_IA32_PERF_CTL	 0x199

// ----------------------------------------------------------------------------------------------
static bool SpeedStep_SetSpeed(uint pos)
{
	unsigned l, h;
	unsigned oldmsr, msr;
	unsigned volt, freq;

	int i = -1;
	while (current_table[i + 1].freq != 0)
		i++;
		
	if (pos > i)
		return false;

	msr = ((( current_table[pos].freq /* MHz */)/100) << 8) | (( current_table[pos].volt /* mV*/ - 700) / 16);

	rdmsr(MSR_IA32_PERF_CTL, oldmsr, h);

	if (msr == (oldmsr & 0xffff))
	{	// No change needed
		return true;
	}

	oldmsr &= ~0xffff;
	msr &= 0xffff;
	oldmsr |= msr;

// - Safety
	volt = oldmsr;
	freq = oldmsr;
	
	volt &=  0xff;
	freq = (freq >> 8) & 0xff;
	
	if (freq * 100 > MAX_FREQUENZY)	// - Too high frequency
		return false;
	if (volt * 16 + 700 > MAX_VOLT)	// - Too high voltage
		return false;
// - Safety End

	wrmsr(MSR_IA32_PERF_CTL, oldmsr, h);
		
	return true;
}

// ----------------------------------------------------------------------------------------------
static bool SpeedStep_Init()
{
	/* Check to see if Enhanced SpeedStep is enabled, and try to
	   enable it if not. */
	unsigned l, h;
	rdmsr(MSR_IA32_MISC_ENABLE, l, h);

	if (!(l & (1<<16)))
	{
		l |= (1<<16);
		wrmsr(MSR_IA32_MISC_ENABLE, l, h);

		rdmsr(MSR_IA32_MISC_ENABLE, l, h);
		if (!(l & (1<<16)))
		{	// Couldn't enable Enhanced SpeedStep
			return false;
		}
	}
	return true;
}

// ----------------------------------------------------------------------------------------------
status_t init_hardware (void)
{
	return B_OK;
}


// ----------------------------------------------------------------------------------------------
status_t init_driver (void)
{
	return B_OK;
}


// ----------------------------------------------------------------------------------------------
void uninit_driver (void)
{
}

	
// ----------------------------------------------------------------------------------------------
static status_t my_device_open (const char *name, uint32 flags, void** cookie)
{
	return B_OK;
}


// ----------------------------------------------------------------------------------------------
static status_t my_device_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	unsigned l, h, msr;
	
	rdmsr(MSR_IA32_PERF_STATUS, l, h);

	*num_bytes = 1;
	msr = l;
	msr = (msr >> 8) & 0xff;

	if (msr == 0)
		rdmsr(MSR_IA32_PERF_CTL, l, h);

	*(unsigned*)buf = l;
	return B_OK;
}

// ----------------------------------------------------------------------------------------------
static status_t my_device_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	*num_bytes = 0;
	return B_IO_ERROR;
}


// ----------------------------------------------------------------------------------------------
static status_t my_device_control (void* cookie, uint32 op, void* arg, size_t len)
{
	switch(op)
	{
		case SPEEDSTEP_INIT:
			return SpeedStep_Init() ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P0:
			return SpeedStep_SetSpeed(0) ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P1:
			return SpeedStep_SetSpeed(1) ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P2:
			return SpeedStep_SetSpeed(2) ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P3:
			return SpeedStep_SetSpeed(3) ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P4:
			return SpeedStep_SetSpeed(4) ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P5:
			return SpeedStep_SetSpeed(5) ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P6:
			return SpeedStep_SetSpeed(6) ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P7:
			return SpeedStep_SetSpeed(7) ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P8:
			return SpeedStep_SetSpeed(8) ? B_OK : B_BAD_VALUE;
		case SPEEDSTEP_SET_P9:
			return SpeedStep_SetSpeed(9) ? B_OK : B_BAD_VALUE;
		default:
			return B_BAD_VALUE;
	}
	return B_BAD_VALUE;
}


// ----------------------------------------------------------------------------------------------
static status_t my_device_close (void* cookie)
{
	return B_OK;
}

// ----------------------------------------------------------------------------------------------
static status_t my_device_free (void* cookie)
{
	return B_OK;
}


// ----------------------------------------------------------------------------------------------
static const char *my_device_name[] =
{
	"misc/speedstep_centrino",
	NULL
};

// ----------------------------------------------------------------------------------------------
device_hooks my_device_hooks = {
	my_device_open, 		/* -> open entry point */
	my_device_close, 		/* -> close entry point */
	my_device_free,			/* -> free cookie */
	my_device_control, 		/* -> control entry point */
	my_device_read,			/* -> read entry point */
	my_device_write			/* -> write entry point */
};

// ----------------------------------------------------------------------------------------------
const char** publish_devices()
{
	return my_device_name;
}

// ----------------------------------------------------------------------------------------------
device_hooks* find_device(const char* name)
{
	return &my_device_hooks;
}

